package net.ihe.gazelle.adapter.connector;

import ca.uhn.fhir.rest.server.exceptions.InvalidRequestException;
import net.ihe.gazelle.app.patientregistryapi.business.EntityIdentifier;
import net.ihe.gazelle.app.patientregistryapi.business.GenderCode;
import net.ihe.gazelle.app.patientregistryapi.business.Person;
import net.ihe.gazelle.app.patientregistryapi.business.PersonName;
import org.hl7.fhir.r4.model.*;
import org.hl7.fhir.r4.model.Address.AddressUse;
import org.hl7.fhir.r4.model.Patient.ContactComponent;

/**
 * Class containing static methods to convert
 *
 * @author pvm
 */
public class BundleToPatientRegistryConverter {

    /**
     * private constructor to override the public implicit one.
     */
    private BundleToPatientRegistryConverter() {
    }



    public static net.ihe.gazelle.app.patientregistryapi.business.Patient fhirPatientToGazellePatient(Patient patient) throws ConversionException {
        net.ihe.gazelle.app.patientregistryapi.business.Patient convertedPatient = new net.ihe.gazelle.app.patientregistryapi.business.Patient();
        if (patient.hasActive()) {
            convertedPatient.setActive(patient.getActive());
        }

        if (patient.hasBirthDate()) {
            convertedPatient.setDateOfBirth(patient.getBirthDate());
        }

        if (patient.hasDeceased()) {
            convertedPatient.setDateOfDeath(patient.getDeceasedDateTimeType().getValue());
        }

        if (patient.hasGender()) {
            convertedPatient.setGender(convertGender(patient.getGender()));
        }

        if (patient.hasMultipleBirth()) {
            convertedPatient.setMultipleBirthOrder(patient.getMultipleBirthIntegerType().getValue());
        }

        if (patient.hasId()) {
            convertedPatient.setUuid(patient.getId());
        }

        convertingLists(patient, convertedPatient);

        return convertedPatient;
    }

    private static void convertingLists(Patient patient, net.ihe.gazelle.app.patientregistryapi.business.Patient convertedPatient) throws ConversionException {
        if (patient.hasAddress()) {
            for (Address address : patient.getAddress()) {
                convertedPatient.addAddress(convertAddress(address));
            }
        }

        if (patient.hasContact()) {
            for (ContactComponent contact : patient.getContact()) {
                convertedPatient.addContact(convertContact(contact));
            }
        }

        if (patient.hasName()) {
            for (HumanName name : patient.getName()) {
                convertedPatient.addName(convertName(name));
            }
        }

        if (patient.hasIdentifier()) {
            for (Identifier id : patient.getIdentifier()) {
                addEntity(convertedPatient, id);
            }
        }
    }

    private static void addEntity(net.ihe.gazelle.app.patientregistryapi.business.Patient convertedPatient, Identifier id) {
        EntityIdentifier entityIdentifier = new EntityIdentifier();

        if (id.hasSystem()) {
            String systemID = id.getSystem();
            systemID = systemID.replace("urn:oid:", "");
            entityIdentifier.setSystemIdentifier(systemID);
            entityIdentifier.setType("ISO");
        }

        if (id.hasValue()) {
            entityIdentifier.setValue(id.getValue());
        }
        else {
            throw new InvalidRequestException("Cannot create Patient without any Identifier");
        }

        convertedPatient.addIdentifier(entityIdentifier);
    }

    private static GenderCode convertGender(Enumerations.AdministrativeGender gender) throws ConversionException {
        if (gender == null) {
            throw new ConversionException("No gender present");
        }

        switch (gender) {
            case MALE:
                return GenderCode.MALE;
            case FEMALE:
                return GenderCode.FEMALE;
            case UNKNOWN:
                return GenderCode.UNDEFINED;
            case OTHER:
                return GenderCode.OTHER;
            default:
                throw new ConversionException(String.format("Cannot map GenderCode : %s", gender));

        }


    }

    private static net.ihe.gazelle.app.patientregistryapi.business.Address convertAddress(Address address) throws ConversionException {
        net.ihe.gazelle.app.patientregistryapi.business.Address convertedAddress = new net.ihe.gazelle.app.patientregistryapi.business.Address();

        if (address.hasCity()) {
            convertedAddress.setCity(address.getCity());
        }

        if (address.hasCountry()) {
            convertedAddress.setCountryIso3(address.getCountry());
        }

        if (address.hasPostalCode()) {
            convertedAddress.setPostalCode(address.getPostalCode());
        }

        if (address.hasState()) {
            convertedAddress.setState(address.getState());
        }

        if (address.hasUse()) {
            convertedAddress.setUse(convertAddressUse(address.getUse()));
        }

        if (address.hasLine()) {
            for (StringType addressLine : address.getLine()) {
                convertedAddress.addLine(addressLine.getValue());
            }
        }
        return convertedAddress;
    }

    private static net.ihe.gazelle.app.patientregistryapi.business.AddressUse convertAddressUse(AddressUse addressUse) throws ConversionException {
        if (addressUse != null) {
            switch (addressUse) {
                case BILLING:
                    return net.ihe.gazelle.app.patientregistryapi.business.AddressUse.BILLING;
                case HOME:
                    return net.ihe.gazelle.app.patientregistryapi.business.AddressUse.HOME;
				case OLD:
					return net.ihe.gazelle.app.patientregistryapi.business.AddressUse.BAD;
				case TEMP:
                    return net.ihe.gazelle.app.patientregistryapi.business.AddressUse.TEMPORARY;
                case WORK:
                    return net.ihe.gazelle.app.patientregistryapi.business.AddressUse.WORK;
                default:
                    throw new ConversionException(String.format("Cannot map address use : %s", addressUse));
            }
        } else {
            throw new ConversionException("No address use present");
        }
    }

    private static Person convertContact(ContactComponent contact) throws ConversionException {
        Person convertedContact = new Person();

        if (contact.hasGender()) {
            convertedContact.setGender(convertGender(contact.getGender()));
        }

        if (contact.hasAddress()) {
            convertedContact.addAddress(convertAddress(contact.getAddress()));
        }

        if (contact.hasName()) {
            convertedContact.addName(convertName(contact.getName()));
        }
        return convertedContact;
    }

    private static PersonName convertName(HumanName name) {
        PersonName convertedName = new PersonName();

        if (name.hasUse()) {
            convertedName.setUse(name.getUse().toString());
        }

        if (name.hasFamily()) {
            convertedName.setFamily(name.getFamily());
        }

        if (name.hasGiven()) {
            for (StringType givenName : name.getGiven()) {
                convertedName.addGiven(givenName.getValue());
            }
        }

        if (name.hasPrefix()) {
            for (StringType prefixName : name.getPrefix()) {
                convertedName.setPrefix(prefixName.getValue());
            }
        }

        if (name.hasSuffix()) {
            for (StringType suffixName : name.getSuffix()) {
                convertedName.setSuffix(suffixName.getValue());
            }
        }

        return convertedName;
    }
}
