package net.ihe.gazelle.adapter.connector;

import net.ihe.gazelle.app.patientregistryapi.business.Address;
import net.ihe.gazelle.app.patientregistryapi.business.ContactPoint;
import net.ihe.gazelle.app.patientregistryapi.business.Patient;
import net.ihe.gazelle.app.patientregistryapi.business.*;
import org.hl7.fhir.exceptions.FHIRException;
import org.hl7.fhir.r4.model.*;
import org.hl7.fhir.r4.model.Bundle.BundleEntryComponent;
import org.hl7.fhir.r4.model.Bundle.BundleType;
import org.hl7.fhir.r4.model.MessageHeader.MessageDestinationComponent;
import org.hl7.fhir.r4.model.MessageHeader.MessageHeaderResponseComponent;
import org.hl7.fhir.r4.model.MessageHeader.MessageSourceComponent;
import org.hl7.fhir.r4.model.MessageHeader.ResponseType;
import org.hl7.fhir.r4.model.Narrative.NarrativeStatus;

import java.util.List;

/**
 * Converter to transform Patient Registry's {@link Patient} to {@link org.hl7.fhir.r4.model.Patient}.
 *
 * @author pvm
 */
public class BusinessToFhirConverter {

    private static final String URN_PREFIX = "urn:oid:";

    /**
     * private constructor to override the implicit one.
     */
    private BusinessToFhirConverter() {
    }

    public static Bundle uuidToBundle(String uuid, String serverUrl) {

        Bundle response = new Bundle();
        response.setId("BundlePIXmResponse");
        Meta metadata = new Meta();
        metadata.addProfile("http://fhir.ch/ig/ch-epr-mhealth/StructureDefinition/ch-pixm-bundle-response");
        response.setType(BundleType.MESSAGE);
        BundleEntryComponent entry = new BundleEntryComponent();
        entry.setFullUrl(serverUrl + uuid);
        MessageHeader messageHeader = new MessageHeader();
        messageHeader.setId(uuid);
        Narrative narrative = new Narrative();
        narrative.setStatus(NarrativeStatus.GENERATED);
        messageHeader.setText(narrative);
        MessageDestinationComponent mdc = new MessageDestinationComponent();
        mdc.setEndpoint(serverUrl + uuid);
        messageHeader.addDestination(mdc);
        MessageSourceComponent msc = new MessageSourceComponent();
        msc.setEndpoint(serverUrl + uuid);
        messageHeader.setSource(msc);
        MessageHeaderResponseComponent mhrc = new MessageHeaderResponseComponent();
        mhrc.setIdentifier(uuid);
        mhrc.setCode(ResponseType.OK);
        messageHeader.setResponse(mhrc);
        entry.setResource(messageHeader);
        response.addEntry(entry);
        response.setMeta(metadata);

        return response;

    }

    public static org.hl7.fhir.r4.model.Patient patientToFhirPatient(Patient patient) throws ConversionException {
        org.hl7.fhir.r4.model.Patient fhirPatient = new org.hl7.fhir.r4.model.Patient();
        fhirPatient.setId(patient.getUuid());
        setNames(fhirPatient, patient);
        fhirPatient.setGender(getGenderCode(patient));
        setCrossIdentifier(fhirPatient, patient);
        setBirthDate(fhirPatient, patient);
        setAddresses(fhirPatient, patient);
        setTelecom(fhirPatient, patient);
        fhirPatient.setActive(patient.isActive());
        return fhirPatient;

    }

    private static void setNames(org.hl7.fhir.r4.model.Patient fhirPatient, Patient patient) throws ConversionException {
        List<PersonName> patientNames = patient.getNames();
        if (patientNames != null) {
            for (PersonName personName : patientNames) {
                if (personName != null) {
                    HumanName name = new HumanName();
                    name.setFamily(personName.getFamily());
                    for (String given : personName.getGivens()) {
                        name.addGiven(given);
                    }
                    try {
                        name.setUse(HumanName.NameUse.fromCode(personName.getUse()));
                    } catch (FHIRException e) {
                        throw new ConversionException(String.format("Cannot convert PersonName use : %s", personName.getUse()), e);
                    }
                    name.addPrefix(personName.getPrefix());
                    name.addSuffix(personName.getSuffix());
                    fhirPatient.addName(name);
                }
            }
        }
    }

    private static Enumerations.AdministrativeGender getGenderCode(Patient patient) throws ConversionException {
        if (patient.getGender() != null) {
            switch (patient.getGender()) {
                case MALE:
                    return Enumerations.AdministrativeGender.MALE;
                case FEMALE:
                    return Enumerations.AdministrativeGender.FEMALE;
                case UNDEFINED:
                    return Enumerations.AdministrativeGender.UNKNOWN;
                case OTHER:
                    return Enumerations.AdministrativeGender.OTHER;
                default:
                    throw new ConversionException(String.format("Cannot map GenderCode : %s", patient.getGender()));
            }
        }
        return null;

    }

    private static void setCrossIdentifier(org.hl7.fhir.r4.model.Patient fhirPatient, Patient patient) {
        if (patient.getIdentifiers() != null) {
            for (EntityIdentifier currentPatientIdentifier : patient.getIdentifiers()) {
                if (currentPatientIdentifier.getSystemIdentifier() != null) {
                    String fhirSystem = getUniversalIDAsUrn(currentPatientIdentifier.getSystemIdentifier());
                    fhirPatient.addIdentifier().setSystem(fhirSystem).setValue(currentPatientIdentifier.getValue());
                }
            }
        }
    }

    private static void setBirthDate(org.hl7.fhir.r4.model.Patient fhirPatient, Patient patient) {
        if (patient.getDateOfBirth() != null) {
            fhirPatient.setBirthDate(patient.getDateOfBirth());
        }
    }

    private static void setAddresses(org.hl7.fhir.r4.model.Patient fhirPatient, Patient patient) throws ConversionException {
        List<Address> addressList = patient.getAddresses();
        if (addressList != null) {
            for (Address address : addressList) {
                org.hl7.fhir.r4.model.Address fhirAddress = new org.hl7.fhir.r4.model.Address();
                for (String line : address.getLines()) {
                    fhirAddress.addLine(line);
                }
                fhirAddress.setCity(address.getCity());
                fhirAddress.setCountry(address.getCountryIso3());
                fhirAddress.setPostalCode(address.getPostalCode());
                fhirAddress.setState(address.getState());
                fhirAddress.setUse(getAddressUse(address.getUse()));
                fhirPatient.addAddress(fhirAddress);
            }
        }
    }

    private static void setTelecom(org.hl7.fhir.r4.model.Patient fhirPatient, Patient patient) throws ConversionException {
        List<ContactPoint> contactPoints = patient.getContactPoints();
        if (contactPoints != null) {
            for (ContactPoint contactPoint : contactPoints) {
                if (contactPoint != null) {
                    org.hl7.fhir.r4.model.ContactPoint fhirContactPoint = new org.hl7.fhir.r4.model.ContactPoint();
                    fhirContactPoint.setSystem(getContactPointSystem(contactPoint.getType()));
                    fhirContactPoint.setValue(contactPoint.getValue());
                    fhirContactPoint.setUse(getContactPointUse(contactPoint.getUse()));
                    fhirPatient.addTelecom(fhirContactPoint);
                }
            }
        }
    }

    private static String getUniversalIDAsUrn(String universalID) {
        if (universalID == null) {
            return null;
        } else if (universalID.startsWith(URN_PREFIX)) {
            return universalID;
        } else {
            return URN_PREFIX + universalID;
        }
    }

    private static org.hl7.fhir.r4.model.Address.AddressUse getAddressUse(AddressUse addressUse) throws ConversionException {
        if (addressUse == null) {
            return null;
        }

        switch (addressUse) {
            case HOME:
            case PRIMARY_HOME:
                return org.hl7.fhir.r4.model.Address.AddressUse.HOME;
            case WORK:
                return org.hl7.fhir.r4.model.Address.AddressUse.WORK;
            case VACATION_HOME:
            case TEMPORARY:
                return org.hl7.fhir.r4.model.Address.AddressUse.TEMP;
            case BAD:
                return org.hl7.fhir.r4.model.Address.AddressUse.OLD;
            case BILLING:
                return org.hl7.fhir.r4.model.Address.AddressUse.BILLING;
            default:
                throw new ConversionException(String.format("Cannot map AddressUse : %s", addressUse));
        }

    }

    private static org.hl7.fhir.r4.model.ContactPoint.ContactPointUse getContactPointUse(ContactPointUse contactPointUse) throws ConversionException {
        if (contactPointUse == null) {
            return null;
        }
        switch (contactPointUse) {
            case HOME:
            case PRIMARY_HOME:
                return org.hl7.fhir.r4.model.ContactPoint.ContactPointUse.HOME;
            case MOBILE:
                return org.hl7.fhir.r4.model.ContactPoint.ContactPointUse.MOBILE;
            case WORK:
                return org.hl7.fhir.r4.model.ContactPoint.ContactPointUse.WORK;
            case TEMPORARY:
                return org.hl7.fhir.r4.model.ContactPoint.ContactPointUse.TEMP;
            default:
                throw new ConversionException(String.format("Cannot map ContactPointUse : %s", contactPointUse.value()));
        }
    }

    private static org.hl7.fhir.r4.model.ContactPoint.ContactPointSystem getContactPointSystem(ContactPointType contactPointType)
            throws ConversionException {
        if (contactPointType == null) {
            return null;
        }
        switch (contactPointType) {
            case BEEPER:
                return org.hl7.fhir.r4.model.ContactPoint.ContactPointSystem.PAGER;
            case PHONE:
                return org.hl7.fhir.r4.model.ContactPoint.ContactPointSystem.PHONE;
            case FAX:
                return org.hl7.fhir.r4.model.ContactPoint.ContactPointSystem.FAX;
            case URL:
                return org.hl7.fhir.r4.model.ContactPoint.ContactPointSystem.URL;
            case EMAIL:
                return org.hl7.fhir.r4.model.ContactPoint.ContactPointSystem.EMAIL;
            case SMS:
                return org.hl7.fhir.r4.model.ContactPoint.ContactPointSystem.SMS;
            case OTHER:
                return org.hl7.fhir.r4.model.ContactPoint.ContactPointSystem.OTHER;
            default:
                throw new ConversionException(String.format("Cannot map ContactPointType : %s", contactPointType));
        }

    }

}
