package net.ihe.gazelle.application;

import ca.uhn.fhir.rest.server.exceptions.*;
import net.ihe.gazelle.adapter.connector.BusinessToFhirConverter;
import net.ihe.gazelle.adapter.connector.DeletionBundle;
import net.ihe.gazelle.adapter.connector.UpdateBundle;
import net.ihe.gazelle.adapter.preferences.Preferences;
import net.ihe.gazelle.app.patientregistryapi.application.PatientFeedException;
import net.ihe.gazelle.app.patientregistryapi.business.PersonName;
import net.ihe.gazelle.app.patientregistryfeedclient.adapter.PatientFeedClient;
import net.ihe.gazelle.framework.loggerservice.application.GazelleLogger;
import net.ihe.gazelle.framework.loggerservice.application.GazelleLoggerFactory;
import net.ihe.gazelle.framework.preferencesmodelapi.application.NamespaceException;
import net.ihe.gazelle.framework.preferencesmodelapi.application.OperationalPreferencesService;
import net.ihe.gazelle.framework.preferencesmodelapi.application.PreferenceException;
import net.ihe.gazelle.lib.annotations.Package;
import org.hl7.fhir.r4.model.Bundle;

import javax.inject.Inject;
import javax.inject.Named;
import javax.xml.ws.WebServiceException;
import java.net.InetAddress;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.UnknownHostException;

@Named("PatientRegistryFeedClient")
public class PatientRegistryFeedClient {

	public static final String NO_UUID = "No UUID was retrieved from the created Patient";
	public static final String NO_PATIENT_PARAMETER = "No Patient were given to create";
	public static final String CLIENT_NOT_SET = "client not set";
	public static final String INVALID_PARAMETERS = "Invalid parameters";
	private String serverUrl;
	private static final GazelleLogger logger = GazelleLoggerFactory.getInstance().getLogger(PatientRegistryFeedClient.class);
	private OperationalPreferencesService operationalPreferencesService;
	private PatientFeedClient client;

	public PatientRegistryFeedClient() {
	}

	/**
	 * Default constructor, used for injection.
	 *
	 * @param operationalPreferencesService {@link OperationalPreferencesService} used to retrieve Patient Registry's URL.
	 */
	@Inject
	public PatientRegistryFeedClient(OperationalPreferencesService operationalPreferencesService) {

		this.operationalPreferencesService = operationalPreferencesService;
	}

	/**
	 * Package-private constructor used for test purposes.
	 *
	 * @param url {@link URL} to be used by this to instantiate the processing service to retrieve patients.
	 */
	@Package
	PatientRegistryFeedClient(URL url) {
		this.client = new PatientFeedClient(url);
	}

	@Package
	public void setClient(PatientFeedClient client) {
		this.client = client;
	}

	/**
	 * Initialize the Search Client using the Operational Preferences Service.
	 *
	 * @throws PatientFeedException if the service cannot be instantiated.
	 */
	private void initializeClient() throws PatientFeedException {
		String patientRegistryUrl = null;
		getUrlOfServer();
		try {
			patientRegistryUrl = this.operationalPreferencesService.
					getStringValue(Preferences.PATIENT_REGISTRY_URL.getNamespace().getValue(), Preferences.PATIENT_REGISTRY_URL.getKey());
			this.client = new PatientFeedClient(new URL(patientRegistryUrl));
		} catch (NamespaceException | PreferenceException e) {
			throw new PatientFeedException(String.format("Unable to retrieve [%s] Preference !", Preferences.PATIENT_REGISTRY_URL.getKey()));
		} catch (MalformedURLException e) {
			throw new PatientFeedException(String.format("Preference [%s] with value [%s] is not a valid URL !",
					Preferences.PATIENT_REGISTRY_URL.getKey(),
					patientRegistryUrl));
		} catch (WebServiceException e) {
			logger.warn(e.getMessage());
			throw new PatientFeedException(String.format("Can't connect to patient registry ! at address [%s]", patientRegistryUrl));
		}
	}

	/**
	 * Method called to create a Patient in the Patient Registry Database
	 *
	 * @param patient : the patient to add in database, represented in patientRegistry model
	 * @return a string corresponding to the uuid of the newly created patient in the database.
	 */
	public Bundle createPatient(net.ihe.gazelle.app.patientregistryapi.business.Patient patient) throws PatientFeedException {
		if (client == null) {
			logger.info(CLIENT_NOT_SET);
			initializeClient();
		}
		if (patient == null) {
			throw new InvalidRequestException(NO_PATIENT_PARAMETER);
		}

		PersonName name = patient.getNames().get(0);
		if (name.getFamily().isBlank() || name.getGivens().isEmpty()) {
			throw new InvalidRequestException("Mandatory fields are missing");
		}
		if (patient.getDateOfBirth() == null || patient.getGender() == null) {
			throw new InvalidRequestException("Mandatory fields are missing");
		}
		String uuid = "";
		try {
			patient.setUuid("PatientPIXmFeed");
			patient.setActive(true);
			uuid = client.createPatient(patient);

			if (uuid == null || uuid.isBlank()) {
				throw new InternalErrorException(NO_UUID);
			}


		} catch (PatientFeedException e) {
			switch (e.getCause().getMessage()) {
				case "Impossible to cross reference the patient (not saved)":
					throw new InternalErrorException("Impossible to cross reference the patient, it will not be saved !", e);
				case "Unexpected Exception persisting Patient !":
					throw new InternalErrorException("Unexpected Exception persisting Patient !", e);

				default:
					treatClientBaseErrors(e);
			}
		}
		return BusinessToFhirConverter.uuidToBundle(uuid, serverUrl);
	}

	/**
	 * Method called to update a Patient in the PAtient Registry Database.
	 *
	 * @param patient : the Patient object we want to insert in DB
	 * @param uuid    : The uuid of the patient corresponding to it in DB
	 * @return a String corresponding to the uuid confirming the transaction has been successful.
	 */
	public Bundle updatePatient(net.ihe.gazelle.app.patientregistryapi.business.Patient patient, String uuid) throws PatientFeedException {
		if (client == null) {
			logger.info(CLIENT_NOT_SET);
			initializeClient();
		}
		if (patient == null || uuid == null || uuid.isBlank()) {
			throw new InvalidRequestException(INVALID_PARAMETERS);
		}

		try {
			patient.setUuid(uuid);
			client.updatePatient(patient);
		} catch (PatientFeedException e) {
			treatClientBaseErrors(e);
		}
		return UpdateBundle.prepareBundleAfterUpdateOperation(uuid,serverUrl);

	}

	/**
	 * Method called to merge two Patients in the Patient Registry Database.
	 *
	 * @param uuidOriginal   : the Patient object we want to insert in DB
	 * @param uuidDuplicated : The uuid of the patient corresponding to it in DB
	 * @return a String corresponding to the uuid confirming the transaction has been successful.
	 */
	public Bundle mergePatient(String uuidOriginal, String uuidDuplicated) throws PatientFeedException {
		if (client == null) {
			logger.info(CLIENT_NOT_SET);
			initializeClient();
		}
		if (uuidOriginal == null || uuidOriginal.isBlank()) {
			throw new InvalidRequestException(INVALID_PARAMETERS);
		}
		if (uuidDuplicated == null || uuidDuplicated.isBlank()) {
			throw new InvalidRequestException(INVALID_PARAMETERS);
		}
		Bundle response = new Bundle();
		try {

			response.setId(client.mergePatient(uuidOriginal, uuidDuplicated));

		} catch (PatientFeedException e) {
			treatClientBaseErrors(e);
		}
		return response;
	}

	/**
	 *  Methode to delete a Patient with its uuid
	 * @param uuid of the patient to delete
	 * @return Result of the Deletion
	 * @throws PatientFeedException
	 */
	public Bundle delete(String uuid) throws PatientFeedException {
		if (client == null) {
			logger.info(CLIENT_NOT_SET);
			initializeClient();
		}
		if (uuid == null || uuid.isBlank()) {
			throw new InvalidRequestException("Invalid parameter");
		}
		try {
			boolean deletionStatus = client.deletePatient(uuid);
			if (!deletionStatus) {
				throw new ResourceGoneException("Patient with UUID " + uuid);
			}
		} catch (PatientFeedException exception) {
			switch (exception.getCause().getMessage()) {
				case "The uuid cannot be null or empty":
					throw new InternalErrorException("The uuid cannot be null or empty", exception);
				case "Cannot proceed to delete":
					throw new InternalErrorException("Cannot proceed to delete", exception);
				default:
					treatClientBaseErrors(exception);
			}
		}
		return DeletionBundle.prepareBundleAfterDeletionOperation(uuid, serverUrl);
	}

	/**
	 * The method is used to retrieve the url of the current server.
	 */
	private void getUrlOfServer() {
		InetAddress ip;
		try {
			ip = InetAddress.getLocalHost();
			serverUrl = ip.getCanonicalHostName();
			logger.info(String.format("Your current Hostname : %s", serverUrl));

		} catch (UnknownHostException exception) {
			logger.error("Unable to find serverUrl");
			serverUrl = "ReplaceByTheRightEndpoint";
		}
	}

	/**
	 * Private Method which goal is to process the first levels exceptions thrown by the Patient feed client.
	 *
	 * @param e the exception thrown
	 * @throws InternalErrorException throws back the exception with the right http code and the stack trace.
	 */
	private void treatClientBaseErrors (Exception e) {
		switch(e.getMessage()) {
			case "Exception while Mapping with GITB elements !":
				throw new InternalErrorException("Exception while Mapping with GITB elements !", e);
			case "Invalid Response from distant PatientFeedProcessingService !":
				throw new InternalErrorException("Invalid Response from distant PatientFeedProcessingService !", e);
			case "Invalid operation used on distant PatientFeedProcessingService !":
				throw new InternalErrorException("Invalid operation used on distant PatientFeedProcessingService !", e);
			case "Invalid Request sent to distant PatientFeedProcessingService !":
				throw new InternalErrorException("Invalid Request sent to distant PatientFeedProcessingService !", e);
			default:
				throw new InternalErrorException("An unhandled error was thrown.", e);
		}
	}
}
