package net.ihe.gazelle.business.provider;

import ca.uhn.fhir.rest.annotation.*;
import ca.uhn.fhir.rest.api.MethodOutcome;
import ca.uhn.fhir.rest.server.IResourceProvider;
import ca.uhn.fhir.rest.server.exceptions.InternalErrorException;
import ca.uhn.fhir.rest.server.exceptions.InvalidRequestException;
import net.ihe.gazelle.application.PatientRegistryFeedClient;
import net.ihe.gazelle.lib.annotations.Package;
import org.hl7.fhir.instance.model.api.IBaseResource;
import org.hl7.fhir.r4.model.Bundle;
import org.hl7.fhir.r4.model.IdType;
import org.hl7.fhir.r4.model.Patient;
import org.hl7.fhir.r4.model.ResourceType;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.inject.Inject;
import javax.inject.Named;
import java.util.List;

@Named("chBundleProvider")
public class CHBundleProvider implements IResourceProvider {
	
    private static final Logger patientLogger = LoggerFactory.getLogger(CHBundleProvider.class);

    private CHBundleProvider() {
    }

    @Package
    public CHBundleProvider(ChPatientResourceProvider chPatientResourceProvider) {
        this.chPatientResourceProvider = chPatientResourceProvider;
    }

    @Inject
    private ChPatientResourceProvider chPatientResourceProvider;

    @Inject
    private PatientRegistryFeedClient patientRegistryFeedClient;

    @Package
    public CHBundleProvider(PatientRegistryFeedClient client) {
        this.patientRegistryFeedClient = client;
    }

    @Override
    public Class<? extends IBaseResource> getResourceType() {
        return Bundle.class;
    }

    /**
     * Method called to create a Patient contained in an iti 93 bundle
     *
     * @param iti93Bundle : the Bundle containing the patient to add to Patient Registry.
     * @return a Method Outcome containing
     */
    @Create
    public MethodOutcome create(@ResourceParam Bundle iti93Bundle) {
        if (iti93Bundle == null || iti93Bundle.getEntry().isEmpty()) {
            patientLogger.error("Bundle is null or Empty");
            throw new InvalidRequestException("Bundle is null or Empty");
        }
        Patient patientBundle = iti93BundleToPatient(iti93Bundle);
        return chPatientResourceProvider.create(patientBundle);
    }

	/**
	 * Method to delete a Patient - represented by its uuid - and a bundle representing the Patientas well.
	 *
     * @param theId : the uuid of the patient to delete in database
	 * @param iti93Bundle : the bundle containing the patient to delete.
     *
	 * @return a method outcome containing a response Bundle with the patient and the http response.
	 */
	@Delete
	public MethodOutcome delete(@IdParam IdType theId, @ResourceParam Bundle iti93Bundle) {

	    if (theId == null || theId.isEmpty()) {
            throw new InvalidRequestException("Invalid ID Parameter, either null or empty.");
        }

		if (iti93Bundle == null || iti93Bundle.getEntry().isEmpty()) {
			throw new InvalidRequestException("Invalid bundle provided, either null or empty.");
		}
		try {
			iti93BundleToPatient(iti93Bundle);
			return chPatientResourceProvider.delete(theId);
		} catch (InvalidRequestException e) {
			throw new InvalidRequestException("Bundle Could not be converted to HL7 Patient");
		}
	}

    /**
     * Method called to update a Patient - represented by its uuid - with another Patient included in an iti 93 Bundle.
     *
     * @param theId : the id of the Patient to update in the database
     * @param iti93Bundle : the new value of the patient to be saved.
     * @return a Method Outcome containing the response bundle with a representation of the patient, and the http response
     */
	@Update
    public MethodOutcome updatePatient(@IdParam IdType theId, @ResourceParam Bundle iti93Bundle) {

        if (theId == null || theId.isEmpty()) {
            throw new InvalidRequestException("Invalid ID Parameter, either null or empty.");
        }

        if (iti93Bundle == null || iti93Bundle.getEntry().isEmpty()) {
            throw new InvalidRequestException("Invalid bundle provided, either null or empty.");
        }

        try {
            Patient updatedPatient = iti93BundleToPatient(iti93Bundle);
            return chPatientResourceProvider.update(theId, updatedPatient);
        } catch (InvalidRequestException e) {
            if (e.getMessage().equals("Cannot find resource Type Patient in Bundle")) {
               throw e;
            } else {
                throw new InvalidRequestException("Patient Resource Provider could not take request.", e);
            }
        } catch (InternalErrorException e) {
            throw new InternalErrorException("Patient Resource Provider could not take request.", e);
        }

    }

    /**
     * Private method to get back the Patient contained in an iti 93 Bundle.
     * @param pixmFeed : the bundle in which to recover the Patient.
     * @return the patient contained in the bundle for further checks and use.
     */
    private Patient iti93BundleToPatient(Bundle pixmFeed) {
        List<Bundle.BundleEntryComponent> listOfEntries = pixmFeed.getEntry();
        if (listOfEntries.size() != 2) {
            throw new InvalidRequestException("PixM Feed should have 2 entries.");
        }
        for (Bundle.BundleEntryComponent entry : listOfEntries) {
            if (entry.getResource().getResourceType() == ResourceType.Bundle) {
                Bundle historyBundle = (Bundle) entry.getResource();
                for (Bundle.BundleEntryComponent resource : historyBundle.getEntry()) {
                    if (resource.getResource().getResourceType() == ResourceType.Patient) {
                        return (Patient) resource.getResource();
                    }
                }
            }
        }
        throw new InvalidRequestException("Cannot find resource Type Patient in Bundle");
    }
}