package net.ihe.gazelle.business.provider;

import ca.uhn.fhir.rest.annotation.*;
import ca.uhn.fhir.rest.api.MethodOutcome;
import ca.uhn.fhir.rest.param.StringAndListParam;
import ca.uhn.fhir.rest.param.StringOrListParam;
import ca.uhn.fhir.rest.param.StringParam;
import ca.uhn.fhir.rest.param.TokenParam;
import ca.uhn.fhir.rest.server.IResourceProvider;
import ca.uhn.fhir.rest.server.exceptions.ForbiddenOperationException;
import ca.uhn.fhir.rest.server.exceptions.InternalErrorException;
import ca.uhn.fhir.rest.server.exceptions.InvalidRequestException;
import ca.uhn.fhir.rest.server.exceptions.ResourceNotFoundException;
import net.ihe.gazelle.adapter.connector.BundleToPatientRegistryConverter;
import net.ihe.gazelle.adapter.connector.ConversionException;
import net.ihe.gazelle.app.patientregistryapi.application.PatientFeedException;
import net.ihe.gazelle.app.patientregistryapi.application.SearchCrossReferenceException;
import net.ihe.gazelle.app.patientregistryapi.business.EntityIdentifier;
import net.ihe.gazelle.application.PatientRegistryFeedClient;
import net.ihe.gazelle.application.PatientRegistrySearchClient;
import net.ihe.gazelle.application.PatientRegistryXRefSearchClient;
import net.ihe.gazelle.lib.annotations.Package;
import net.ihe.gazelle.lib.searchmodelapi.business.exception.SearchException;
import org.hl7.fhir.instance.model.api.IBaseResource;
import org.hl7.fhir.r4.model.IdType;
import org.hl7.fhir.r4.model.Parameters;
import org.hl7.fhir.r4.model.Patient;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.inject.Inject;
import javax.inject.Named;
import java.util.ArrayList;
import java.util.List;

/**
 * This is a resource provider which stores Patient resources in memory using a HashMap. This is obviously not a production-ready solution for many
 * reasons,
 * but it is useful to help illustrate how to build a fully-functional server.
 */
@Named("chPatientResourceProvider")
public class ChPatientResourceProvider implements IResourceProvider {

    public static final String URN_OID = "urn:oid:";
    public static final String INVALID_REQUEST_BAD_SOURCE_IDENTIFIER = "Invalid request : bad sourceIdentifier";
    public static final String INVALID_REQUEST_TARGET_DOMAIN_CAN_NOT_BE_EMPTY = "Invalid request : targetDomain can not be empty";
    public static final String INVALID_REQUEST_THE_REQUEST_SHALL_CONTAIN_ONE_OR_TWO_TARGET_DOMAIN = "Invalid request : The request shall contain " +
            "one or two targetDomain";
    public static final String SOURCE_IDENTIFIER_PATIENT_IDENTIFIER_NOT_FOUND = "sourceIdentifier Patient Identifier not found";
    public static final String SOURCE_IDENTIFIER_ASSIGNING_AUTHORITY_NOT_FOUND = "sourceIdentifier Assigning Authority not found";
    public static final String TARGET_SYSTEM_NOT_FOUND = "targetSystem not found";
    private static final Logger patientLogger = LoggerFactory.getLogger(ChPatientResourceProvider.class);
    public static final String NO_ID_PROVIDED = "No ID provided";
    public static final String NO_BUNDLE_PROVIDED = "No Bundle provided";
    public static final String NO_PATIENT_PROVIDED = "No Patient Provided";

    @Inject
    private PatientRegistryXRefSearchClient patientRegistryXRefSearchClient;

    @Inject
    private PatientRegistrySearchClient patientRegistrySearchClient;

    @Inject
    private PatientRegistryFeedClient patientRegistryFeedClient;

    private ChPatientResourceProvider() {
    }

    @Package
    public ChPatientResourceProvider(PatientRegistryXRefSearchClient client) {
        this.patientRegistryXRefSearchClient = client;
    }

    @Package
    public ChPatientResourceProvider(PatientRegistryFeedClient client) {
        this.patientRegistryFeedClient = client;
    }

    @Package
    public ChPatientResourceProvider(PatientRegistrySearchClient client) {
        this.patientRegistrySearchClient = client;
    }


    /**
     * The getResourceType method comes from IResourceProvider, and must be overridden to indicate what type of resource this provider supplies.
     */
    @Override
    public Class<? extends IBaseResource> getResourceType() {
        return Patient.class;
    }

    /**
     * Method to read a Patient through its uuid in the patient Manager database
     *
     * @param theId : Id Type given in the url. represents the uuid of the searched patient
     * @return One patient corresponding to the uuid given in entry. If many Patients are found, we consider it as an error and won't return ANY
     * patient.
     */
    @Read
    public Patient read(@IdParam IdType theId) {
        String uuid = theId.getIdPart();
        if (uuid == null || uuid.isBlank()) {
            patientLogger.error(NO_ID_PROVIDED);
            throw new InvalidRequestException(NO_ID_PROVIDED);
        }

        try {
            Patient retrievedPatient = patientRegistrySearchClient.searchPatient(uuid);
            patientLogger.info("Patient Successfully found");
            return retrievedPatient;
        } catch (SearchException e) {
            throw new InternalErrorException("Patient could not be retrieved", e);
        }

    }

    /**
     * Method called to create a Patient on Patient Registry
     *
     * @param iti93Patient : the Patient to create.
     * @return a MethodOutcome containing a representation of the patient created.
     */
    @Create
    public MethodOutcome create(@ResourceParam Patient iti93Patient) {
        if (iti93Patient == null) {
            patientLogger.error(NO_PATIENT_PROVIDED);
            throw new InvalidRequestException(NO_PATIENT_PROVIDED);
        }
        try {
            net.ihe.gazelle.app.patientregistryapi.business.Patient patient =
                    BundleToPatientRegistryConverter.fhirPatientToGazellePatient(iti93Patient);
            MethodOutcome methodOutcome = new MethodOutcome();
            methodOutcome.setResource(patientRegistryFeedClient.createPatient(patient));
            if (methodOutcome.getResource() != null) {
                return methodOutcome;
            } else {
                throw new ResourceNotFoundException("MethodOutcome was not created");
            }

        } catch (ConversionException e) {
            throw new InvalidRequestException("Bundle Could not be converted to HL7 Patient");
        } catch (PatientFeedException e) {
            throw new InternalErrorException("Patient Feed client is not set");
        }
    }

    /**
     * Method to Update a patient related to PIXm Delete Method
     *
     * @param theId      the UUID of the Patient we want to update
     * @param hl7Patient The Bundle content of the patient
     * @return FhirBundle that contains the updated patient
     */
    @Update
    public MethodOutcome update(@IdParam IdType theId, @ResourceParam Patient hl7Patient) {
        String uuid = theId.getIdPart();
        if (uuid == null || uuid.isBlank()) {
            patientLogger.error(NO_ID_PROVIDED);
            throw new InvalidRequestException(NO_ID_PROVIDED);
        }
        if (hl7Patient == null || !hl7Patient.hasId()) {
            throw new InvalidRequestException(NO_BUNDLE_PROVIDED);
        }
        try {
            net.ihe.gazelle.app.patientregistryapi.business.Patient patientToUpdate =
                    BundleToPatientRegistryConverter.fhirPatientToGazellePatient(hl7Patient);
            MethodOutcome methodOutcome = new MethodOutcome();
            methodOutcome.setResource(patientRegistryFeedClient.updatePatient(patientToUpdate, uuid));
            return methodOutcome;
        } catch (ConversionException e) {
            throw new InvalidRequestException("Bundle Could not be converted to HL7 Patient", e);
        } catch (PatientFeedException e) {
            throw new InternalErrorException("Patient Feed client is not set");
        }
    }

    /**
     * Method to delete a Patient related to PIXm Delete Method
     *
     * @param theId of the patient to delete
     * @return FhirBundle that contains the deletion status
     */
    @Delete
    public MethodOutcome delete(@IdParam IdType theId) {

        if (theId == null || theId.isEmpty()) {
            throw new InvalidRequestException(NO_ID_PROVIDED);
        }

        String uuid = theId.getIdPart();
        try {
            MethodOutcome methodOutcome = new MethodOutcome();
            methodOutcome.setResource(patientRegistryFeedClient.delete(uuid));
            return methodOutcome;
        } catch (PatientFeedException e) {
            patientLogger.error("Error in deletion :", e);
            throw new InternalErrorException("Internal Error :", e);
        }
    }

    /**
     * Search method for a Patient using the source identifier required parameter
     * and an optional list of target system
     *
     * @param sourceIdentifierParam : the source identifier of the patient, should be formatted "urn:oid:x.x.x.x.x.x.x.x.x.x|value"
     * @param targetSystemParam:    the target System(s) we want to find the Patient on, should be formatted "urn:oid:x.x.x.x.x.x.x.x.x.x,urn:oid:x
     *                              .x.x.x.x.x.x.x.x.x,"
     * @return a Parameter element composed of a list of target identifier for every Patient found, and an url to the Patient in the server.
     */
    @Operation(name = "$ihe-pix", idempotent = true)
    public Parameters findPatientsByIdentifier(@OperationParam(name = "sourceIdentifier", min = 1, max = 1) TokenParam sourceIdentifierParam,
                                               @OperationParam(name = "targetSystem", min = 1, max = 2) StringAndListParam targetSystemParam) {

        EntityIdentifier sourceIdentifier = createEntityIdentifierFromSourceIdentifier(sourceIdentifierParam);
        List<String> targetSystemList = createTargetSystemListFromParam(targetSystemParam);

        Parameters parametersResults = null;
        try {
            parametersResults = patientRegistryXRefSearchClient.process(sourceIdentifier, targetSystemList);
        } catch (SearchCrossReferenceException e) {
            throw new InternalErrorException(e);
        }
        return parametersResults;
    }

    /**
     * Private method used to transform a Source Identifier into an Entity Identifier for Patient Registry database
     * @param sourceIdentifier : the source Identifier to convert
     * @return An Entity Identifier understood by Patient Registry.
     */
    private EntityIdentifier createEntityIdentifierFromSourceIdentifier(TokenParam sourceIdentifier) {
        if (sourceIdentifier == null) {
            throw new InvalidRequestException(SOURCE_IDENTIFIER_PATIENT_IDENTIFIER_NOT_FOUND);
        }
        String sourceIdentifierSystem = sourceIdentifier.getSystem();
        String sourceIdentifierValue = sourceIdentifier.getValue();
        if (sourceIdentifierSystem == null) {
            patientLogger.error(INVALID_REQUEST_BAD_SOURCE_IDENTIFIER + " null system");
            throw new InvalidRequestException(SOURCE_IDENTIFIER_PATIENT_IDENTIFIER_NOT_FOUND);
        }
        if (sourceIdentifierValue == null) {
            patientLogger.error(INVALID_REQUEST_BAD_SOURCE_IDENTIFIER + " null value");
            throw new InvalidRequestException(SOURCE_IDENTIFIER_ASSIGNING_AUTHORITY_NOT_FOUND);
        }
        if (sourceIdentifierSystem.isBlank() || sourceIdentifierValue.isBlank()) {
            patientLogger.error(INVALID_REQUEST_BAD_SOURCE_IDENTIFIER + " empty system or value");
            throw new InvalidRequestException(SOURCE_IDENTIFIER_ASSIGNING_AUTHORITY_NOT_FOUND);
        }
        patientLogger.info("Searching Patient with given sourceIdentifier");

        EntityIdentifier wellFormedEntityIdentifier = new EntityIdentifier();
        if (sourceIdentifierSystem.contains(URN_OID)) {
            wellFormedEntityIdentifier.setSystemIdentifier(sourceIdentifierSystem.replace(URN_OID, ""));
            wellFormedEntityIdentifier.setType("ISO");
            wellFormedEntityIdentifier.setValue(sourceIdentifierValue);
        } else if (sourceIdentifierSystem.contains("http")) {
            wellFormedEntityIdentifier.setSystemIdentifier(sourceIdentifierSystem);
            wellFormedEntityIdentifier.setType("ISO");
            wellFormedEntityIdentifier.setValue(sourceIdentifierValue);
        } else {
            throw new ResourceNotFoundException(SOURCE_IDENTIFIER_PATIENT_IDENTIFIER_NOT_FOUND);
        }
        return wellFormedEntityIdentifier;
    }

    /**
     * Create the targetSystem list for CH Implementation
     *
     * @param targetSystemParam Param from the query
     * @return List of targetSystem
     */
    private List<String> createTargetSystemListFromParam(StringAndListParam targetSystemParam) {
        if (targetSystemParam == null) {
            throw new ForbiddenOperationException(TARGET_SYSTEM_NOT_FOUND);
        }
        List<String> targetSystemList = new ArrayList<>();
        for (StringOrListParam listParam : targetSystemParam.getValuesAsQueryTokens()) {
            List<StringParam> queryStrings = listParam.getValuesAsQueryTokens();
            if (queryStrings.isEmpty() || queryStrings.size() > 2) {
                throw new ForbiddenOperationException(TARGET_SYSTEM_NOT_FOUND);
            }
            for (StringParam singleParam : queryStrings) {
                String singleParamValue = singleParam.getValue();
                if (singleParamValue.contains(URN_OID)) {
                    singleParamValue = singleParamValue.replace(URN_OID, "");
                }
                if (singleParamValue.equals("")) {
                    throw new ForbiddenOperationException(TARGET_SYSTEM_NOT_FOUND);
                }
                targetSystemList.add(singleParamValue);
            }
        }
        if (targetSystemList.size()>2){
            throw new ForbiddenOperationException(TARGET_SYSTEM_NOT_FOUND);
        }
        return targetSystemList;
    }

}


