package net.ihe.gazelle.business.provider;

import java.util.ArrayList;
import java.util.List;

import javax.inject.Inject;
import javax.inject.Named;

import org.hl7.fhir.instance.model.api.IBaseResource;
import org.hl7.fhir.r4.model.IdType;
import org.hl7.fhir.r4.model.Parameters;
import org.hl7.fhir.r4.model.Patient;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import ca.uhn.fhir.rest.annotation.IdParam;
import ca.uhn.fhir.rest.annotation.Operation;
import ca.uhn.fhir.rest.annotation.OperationParam;
import ca.uhn.fhir.rest.annotation.Read;
import ca.uhn.fhir.rest.param.StringAndListParam;
import ca.uhn.fhir.rest.param.StringOrListParam;
import ca.uhn.fhir.rest.param.StringParam;
import ca.uhn.fhir.rest.param.TokenParam;
import ca.uhn.fhir.rest.server.IResourceProvider;
import ca.uhn.fhir.rest.server.exceptions.ForbiddenOperationException;
import ca.uhn.fhir.rest.server.exceptions.InternalErrorException;
import ca.uhn.fhir.rest.server.exceptions.InvalidRequestException;
import ca.uhn.fhir.rest.server.exceptions.ResourceNotFoundException;
import net.ihe.gazelle.app.patientregistryapi.application.SearchCrossReferenceException;
import net.ihe.gazelle.app.patientregistryapi.business.EntityIdentifier;
import net.ihe.gazelle.application.PatientRegistrySearchClient;
import net.ihe.gazelle.application.PatientRegistryXRefSearchClient;
import net.ihe.gazelle.lib.annotations.Package;
import net.ihe.gazelle.lib.searchmodelapi.business.exception.SearchException;

/**
 * This is a resource provider which stores Patient resources in memory using a HashMap. This is obviously not a production-ready solution for many
 * reasons,
 * but it is useful to help illustrate how to build a fully-functional server.
 */
@Named("ihePatientResourceProvider")
public class IhePatientResourceProvider implements IResourceProvider {
	private static final String URN_OID = "urn:oid:";
	public static final String INVALID_REQUEST_BAD_SOURCE_IDENTIFIER = "Invalid request : bad sourceIdentifier";
	private static final Logger patientLogger = LoggerFactory.getLogger(IhePatientResourceProvider.class);
	public static final String SOURCE_IDENTIFIER_PATIENT_IDENTIFIER_NOT_FOUND = "sourceIdentifier Patient Identifier not found";
	public static final String SOURCE_IDENTIFIER_ASSIGNING_AUTHORITY_NOT_FOUND = "sourceIdentifier Assigning Authority not found";
	public static final String TARGET_SYSTEM_NOT_FOUND = "targetSystem not found";
	public static final String NO_ID_PROVIDED = "No ID provided";

	@Inject
	private PatientRegistryXRefSearchClient patientRegistryXRefSearchClient;

	@Inject
	private PatientRegistrySearchClient patientRegistrySearchClient;

	/**
	 * The getResourceType method comes from IResourceProvider, and must be overridden to indicate what type of resource this provider supplies.
	 */
	@Override
	public Class<? extends IBaseResource> getResourceType() {
		return Patient.class;
	}

	public IhePatientResourceProvider() {
	}

	@Package
	public IhePatientResourceProvider(PatientRegistryXRefSearchClient client) {
		this.patientRegistryXRefSearchClient = client;
	}

	@Package
	public IhePatientResourceProvider(PatientRegistrySearchClient client) {
		this.patientRegistrySearchClient = client;
	}

	/**
	 * Method to read a Patient through its uuid in the patient Manager database
	 *
	 * @param theId : Id Type given in the url. represents the uuid of the searched patient
	 * @return One patient corresponding to the uuid given in entry. If many Patients are found, we consider it as an error and won't return ANY
	 * patient.
	 */
	@Read
	public Patient read(@IdParam IdType theId) {
		if (theId.getIdPart() == null || theId.getIdPart().isBlank()) {
			patientLogger.error(NO_ID_PROVIDED);
			throw new InvalidRequestException(NO_ID_PROVIDED);
		}
		String uuid = theId.getIdPart();

		try {
			Patient retrievedPatient = patientRegistrySearchClient.searchPatient(uuid);
			patientLogger.info("Patient Successfully found");
			return retrievedPatient;
		} catch (SearchException e) {
			throw new InternalErrorException("Patient could not be retrieved", e);
		}

	}

	/**
	 * Search method for a Patient using the source identifier required parameter
	 * and an optional list of target system
	 *
	 * @param sourceIdentifierParam : the source identifier of the patient, should be formatted "urn:oid:x.x.x.x.x.x.x.x.x.x|value"
	 * @param targetSystemParam:    the target System(s) we want to find the Patient on, should be formatted "urn:oid:x.x.x.x.x.x.x.x.x.x,
	 * 								urn:oid:x.x.x.x.x.x.x.x.x.x"
	 * @return a Parameters element composed of a list of target identifier for every Patient found, and an url to the Patient in the server.
	 */
	@Operation(name = "$ihe-pix", idempotent = true)
	public Parameters findPatientsByIdentifier(@OperationParam(name = "sourceIdentifier", min = 1, max = 1) TokenParam sourceIdentifierParam,
			@OperationParam(name = "targetSystem") StringAndListParam targetSystemParam) {

		EntityIdentifier sourceIdentifier = createEntityIdentifierFromSourceIdentifier(sourceIdentifierParam);
		List<String> targetSystemList = createTargetSystemListFromParam(targetSystemParam);

		Parameters parametersResults = null;
		try {
			parametersResults = patientRegistryXRefSearchClient.process(sourceIdentifier, targetSystemList);
		} catch (SearchCrossReferenceException e) {
			throw new InternalErrorException(e);
		}

		return parametersResults;
	}

	private EntityIdentifier createEntityIdentifierFromSourceIdentifier(TokenParam sourceIdentifier) {
		if (sourceIdentifier == null) {
			throw new InvalidRequestException(SOURCE_IDENTIFIER_PATIENT_IDENTIFIER_NOT_FOUND);
		}
		String sourceIdentifierSystem = sourceIdentifier.getSystem();
		String sourceIdentifierValue = sourceIdentifier.getValue();
		if (sourceIdentifierSystem == null) {
			patientLogger.error(INVALID_REQUEST_BAD_SOURCE_IDENTIFIER + " null system");
			throw new InvalidRequestException(SOURCE_IDENTIFIER_PATIENT_IDENTIFIER_NOT_FOUND);
		}
		if (sourceIdentifierValue == null) {
			patientLogger.error(INVALID_REQUEST_BAD_SOURCE_IDENTIFIER + " null value");
			throw new InvalidRequestException(SOURCE_IDENTIFIER_ASSIGNING_AUTHORITY_NOT_FOUND);
		}
		if (sourceIdentifierSystem.isBlank() || sourceIdentifierValue.isBlank()) {
			patientLogger.error(INVALID_REQUEST_BAD_SOURCE_IDENTIFIER + " empty system or value");
			throw new InvalidRequestException(SOURCE_IDENTIFIER_ASSIGNING_AUTHORITY_NOT_FOUND);
		}
		patientLogger.info("Searching Patient with given sourceIdentifier");

		EntityIdentifier wellFormedEntityIdentifier = new EntityIdentifier();
		if (sourceIdentifierSystem.contains(URN_OID)) {
			wellFormedEntityIdentifier.setSystemIdentifier(sourceIdentifierSystem.replace(URN_OID, ""));
			wellFormedEntityIdentifier.setType("ISO");
			wellFormedEntityIdentifier.setValue(sourceIdentifierValue);
		} else if (sourceIdentifierSystem.contains("http")) {
			wellFormedEntityIdentifier.setSystemIdentifier(sourceIdentifierSystem);
			wellFormedEntityIdentifier.setType("ISO");
			wellFormedEntityIdentifier.setValue(sourceIdentifierValue);
		} else {
			throw new ResourceNotFoundException(SOURCE_IDENTIFIER_PATIENT_IDENTIFIER_NOT_FOUND);
		}
		return wellFormedEntityIdentifier;
	}

	private List<String> createTargetSystemListFromParam(StringAndListParam targetSystemParam) {
		List<String> targetSystemList = new ArrayList<>();
		if (targetSystemParam != null) {
			for (StringOrListParam listParam : targetSystemParam.getValuesAsQueryTokens()) {
				List<StringParam> queryStrings = listParam.getValuesAsQueryTokens();
				for (StringParam singleParam : queryStrings) {
					String singleParamValue = singleParam.getValue();
					if (singleParamValue.contains(URN_OID)) {
						singleParamValue = singleParamValue.replace(URN_OID, "");
					}
					if (singleParamValue.equals("")) {
						throw new ForbiddenOperationException(TARGET_SYSTEM_NOT_FOUND);
					}
					targetSystemList.add(singleParamValue);
				}
			}
		}
		return targetSystemList;
	}

}
