package net.ihe.gazelle.business.provider;

import ca.uhn.fhir.rest.api.MethodOutcome;
import ca.uhn.fhir.rest.server.exceptions.InvalidRequestException;
import io.qameta.allure.*;
import net.ihe.gazelle.application.PatientRegistryFeedClient;
import org.hl7.fhir.r4.model.IdType;
import org.junit.jupiter.api.BeforeAll;
import org.junit.jupiter.api.Test;

import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertThrows;

@Feature("CHBundleProvider Test")
class CHBundleProviderTest {

    @BeforeAll
    static void initialized() {

    }

    private static PatientRegistryFeedClient feedClientMock = new PatientRegistryFeedClientMock();

    private static final CHBundleProvider bundle_feed_provider = new CHBundleProvider(new CHPatientProviderMock(feedClientMock));
    public static final String GOOD_UUID = "123e4567-e89b-12d3-a456-426614174000";
    public static final String Throw_two_id = "LOLILO";
    public static final String Throw_ID = "MDR";

    CHBundleProviderMock cbpm = new CHBundleProviderMock();


    @Test
    @Description("Test on the Create Method for pixm-connector")
    void testCreatePatient() {
        String fileName = "post_request.json";
        MethodOutcome mo = bundle_feed_provider.create(cbpm.returnBundleFromResource(fileName));
        assertEquals(cbpm.ID_TO_RETURN,mo.getId().getValue());
    }

    @Test
    @Description("Test on the Create Method for pixm-connector with a null Bundle")
    void testCreateNullBundle() {
        String fileName = "null.json";
        assertThrows(InvalidRequestException.class, ()-> bundle_feed_provider.create(cbpm.returnBundleFromResource(fileName)),"Bundle is null or Empty");
    }

    @Test
    @Description("Test on the Create Method with a Bundle with No Patient inside")
    void testCreateNoPatientInBundle() {
        String fileName = "post_request_NO_PATIENT.json";
        assertThrows(InvalidRequestException.class, ()-> bundle_feed_provider.create(cbpm.returnBundleFromResource(fileName)),"Missing Patient in Bundle");
    }

    @Test
    @Description("Test on the Create Method with a Bundle with 3 entries")
    void testCreatePatientBundleWith1Entry() {
        String fileName = "post_request_1_entry.json";
        assertThrows(InvalidRequestException.class, ()-> bundle_feed_provider.create(cbpm.returnBundleFromResource(fileName)));
    }

    @Test
    @Description("Test on the delete operation, clean delete without issue")
    @Severity(SeverityLevel.CRITICAL)
    @Story("delete operation")
    void testDeletePatient() {
        String fileName = "post_request.json";
        IdType id = new IdType(GOOD_UUID);
        MethodOutcome mo = bundle_feed_provider.delete(id, cbpm.returnBundleFromResource(fileName));
        assertEquals(cbpm.ID_TO_RETURN,mo.getId().getValue());
    }

    @Test
    @Description("Test on the delete operation, deleting a Patient but with a null bundle")
    @Severity(SeverityLevel.CRITICAL)
    @Story("delete operation")
    void testDeletePatientNullBundle() {
        String fileName = "null.json";
        IdType id = new IdType(GOOD_UUID);
        assertThrows(InvalidRequestException.class, ()-> bundle_feed_provider.delete(id, cbpm.returnBundleFromResource(fileName)),"Bundle is null or Empty");
    }

    @Test
    @Description("Test on the delete operation, deleting a Patient but there is no Patient in the bundle")
    @Severity(SeverityLevel.CRITICAL)
    @Story("delete operation")
    void testDeletePatientBlank() {
        String fileName = "post_request_NO_PATIENT.json";
        IdType id = new IdType(GOOD_UUID);
        assertThrows(InvalidRequestException.class, ()-> bundle_feed_provider.delete(id, cbpm.returnBundleFromResource(fileName)),"Bundle Could not be converted to HL7 Patient");
    }

    @Test
    @Description("Test on the delete operation, deleting a Patient but with a null bundle")
    @Severity(SeverityLevel.CRITICAL)
    @Story("delete operation")
    void testDeleteBlankId() {
        String fileName = "post_request.json";
        IdType id = new IdType("");
        assertThrows(InvalidRequestException.class, ()-> bundle_feed_provider.delete(id, cbpm.returnBundleFromResource(fileName)),"Bundle is null or Empty");
    }

    @Test
    @Description("Test on the update operation, clean update without issue")
    @Severity(SeverityLevel.CRITICAL)
    @Story("update operation")
    void testUpdatePatient() {
        String fileName = "post_request.json";
        IdType id = new IdType(GOOD_UUID);
        MethodOutcome mo = bundle_feed_provider.updatePatient(id, cbpm.returnBundleFromResource(fileName));
        assertEquals(cbpm.ID_TO_RETURN,mo.getId().getValue());
    }

    @Test
    @Description("Test on the update operation, updating a Patient but with a null bundle")
    @Severity(SeverityLevel.CRITICAL)
    @Story("update operation")
    void testUpdatePatientNullBundle() {
        String fileName = "null.json";
        IdType id = new IdType(GOOD_UUID);
        assertThrows(InvalidRequestException.class, ()-> bundle_feed_provider.updatePatient(id, cbpm.returnBundleFromResource(fileName)),"Bundle is null or Empty");
    }

    @Test
    @Description("Test on the update operation, updating a Patient but there is no Patient in the bundle")
    @Severity(SeverityLevel.CRITICAL)
    @Story("update operation")
    void testUpdatePatientBlank() {
        String fileName = "post_request_NO_PATIENT.json";
        IdType id = new IdType(GOOD_UUID);
        assertThrows(InvalidRequestException.class, ()-> bundle_feed_provider.updatePatient(id, cbpm.returnBundleFromResource(fileName)),"bizu");
    }

    @Test
    @Description("Test on the update operation, updating a Patient but there is no Patient in the bundle")
    @Severity(SeverityLevel.CRITICAL)
    @Story("update operation")
    void testUpdateEmptyId() {
        String fileName = "post_request_NO_PATIENT.json";
        IdType id = new IdType("");
        assertThrows(InvalidRequestException.class, ()-> bundle_feed_provider.updatePatient(id, cbpm.returnBundleFromResource(fileName)),"Invalid ID Parameter, either null or empty.");
    }

    @Test
    @Description("Test on the update operation, updating a Patient but there is no Patient in the bundle")
    @Severity(SeverityLevel.CRITICAL)
    @Story("update operation")
    void testUpdateInvalidRequest() {
        String fileName = "post_request_NO_PATIENT.json";
        IdType id = new IdType(Throw_ID);
        assertThrows(InvalidRequestException.class, ()-> bundle_feed_provider.updatePatient(id, cbpm.returnBundleFromResource(fileName)),"Invalid ID Parameter, either null or empty.");
    }

    @Test
    @Description("Test on the update operation, updating a Patient but there is no Patient in the bundle")
    @Severity(SeverityLevel.CRITICAL)
    @Story("update operation")
    void testUpdateInternalError() {
        String fileName = "post_request_NO_PATIENT.json";
        IdType id = new IdType(Throw_two_id);
        assertThrows(InvalidRequestException.class, ()-> bundle_feed_provider.updatePatient(id, cbpm.returnBundleFromResource(fileName)),"Invalid ID Parameter, either null or empty.");
    }


}