package net.ihe.gazelle.business.provider;

import ca.uhn.fhir.rest.api.MethodOutcome;
import ca.uhn.fhir.rest.param.StringAndListParam;
import ca.uhn.fhir.rest.param.StringOrListParam;
import ca.uhn.fhir.rest.param.StringParam;
import ca.uhn.fhir.rest.param.TokenParam;
import ca.uhn.fhir.rest.server.exceptions.ForbiddenOperationException;
import ca.uhn.fhir.rest.server.exceptions.InternalErrorException;
import ca.uhn.fhir.rest.server.exceptions.InvalidRequestException;
import ca.uhn.fhir.rest.server.exceptions.ResourceNotFoundException;
import io.qameta.allure.*;
import net.ihe.gazelle.application.PatientRegistryFeedClient;
import net.ihe.gazelle.application.PatientRegistrySearchClient;
import net.ihe.gazelle.application.PatientRegistryXRefSearchClient;
import org.hl7.fhir.r4.model.*;
import org.hl7.fhir.r4.model.Parameters.ParametersParameterComponent;
import org.junit.jupiter.api.BeforeAll;
import org.junit.jupiter.api.Test;

import static org.junit.jupiter.api.Assertions.*;

@Feature("ChaPatientProvider")
class ChPatientResourceProviderTest {
	private static PatientRegistryXRefSearchClient xRefSearchClientMock = new PatientRegistryXRefSearchClientMock();
	private static PatientRegistrySearchClient searchClientMock = new PatientRegistrySearchClientMock();
	private static PatientRegistryFeedClient feedClientMock = new PatientRegistryFeedClientMock();
	private static final ChPatientResourceProvider provider = new ChPatientResourceProvider(xRefSearchClientMock);
	private static final ChPatientResourceProvider read_provider = new ChPatientResourceProvider(searchClientMock);
	private static final ChPatientResourceProvider feed_provider = new ChPatientResourceProvider(feedClientMock);

	@BeforeAll
	static void initialized() {
		xRefSearchClientMock = new PatientRegistryXRefSearchClientMock();
	}

	@Test
	@Description("Test on ihe_pix operation, when aiming for a patient on 1 target System, with unr:oid: type system identifiers")
	@Severity(SeverityLevel.CRITICAL)
	@Story("ihe_pix operation")
	void testOkPatient() {
		TokenParam sourceId = new TokenParam("urn:oid:1", "69420");
		StringAndListParam targetDomains = new StringAndListParam();
		StringOrListParam stringParam = new StringOrListParam();
		stringParam.add(new StringParam("urn:oid:2"));
		targetDomains.addAnd(stringParam);

		Parameters response = provider.findPatientsByIdentifier(sourceId, targetDomains);
		ParametersParameterComponent singleResponse = response.getParameter().get(0);
		Identifier r4 = (Identifier) singleResponse.getValue();
		assertEquals(PatientRegistryXRefSearchClientMock.URN_OK, r4.getValue());
		assertEquals(PatientRegistryXRefSearchClientMock.URN_OK, r4.getSystem());

	}

	@Test
	@Description("Test on ihe_pix operation, when aiming for a patient on 1 target System, with http system identifiers")
	@Severity(SeverityLevel.CRITICAL)
	@Story("ihe_pix operation")
	void testOkPatientHttp() {
		TokenParam sourceId = new TokenParam("http://1", "69420");
		StringAndListParam targetDomains = new StringAndListParam();
		StringOrListParam stringParam = new StringOrListParam();
		stringParam.add(new StringParam("http://2"));
		targetDomains.addAnd(stringParam);

		Parameters response = provider.findPatientsByIdentifier(sourceId, targetDomains);
		ParametersParameterComponent singleResponse = response.getParameter().get(0);
		Identifier r4 = (Identifier) singleResponse.getValue();
		assertEquals(PatientRegistryXRefSearchClientMock.HTTP_OK, r4.getValue());
		assertEquals(PatientRegistryXRefSearchClientMock.HTTP_OK, r4.getSystem());

	}

	@Test
	@Description("Test on ihe_pix operation, when aiming for a patient on 2 target System")
	@Severity(SeverityLevel.CRITICAL)
	@Story("ihe_pix operation")
	void testOkPatientTwoTargetSystem() {
		TokenParam sourceId = new TokenParam("http://1", "69420");
		StringAndListParam targetDomains = new StringAndListParam();
		StringOrListParam stringParam = new StringOrListParam();
		stringParam.add(new StringParam("http://2"));
		stringParam.add(new StringParam("http://3"));
		targetDomains.addAnd(stringParam);

		Parameters response = provider.findPatientsByIdentifier(sourceId, targetDomains);
		ParametersParameterComponent singleResponse = response.getParameter().get(0);
		Identifier r4 = (Identifier) singleResponse.getValue();
		assertEquals(PatientRegistryXRefSearchClientMock.HTTP_OK, r4.getValue());
		assertEquals(PatientRegistryXRefSearchClientMock.HTTP_OK, r4.getSystem());

	}

	@Test
	@Description("Test on ihe_pix operation, when the patient source identifiers is null")
	@Severity(SeverityLevel.CRITICAL)
	@Story("ihe_pix operation")
	void testSourceIdentifierNull() {
		TokenParam sourceId = new TokenParam();
		StringAndListParam targetDomains = new StringAndListParam();
		StringOrListParam stringParam = new StringOrListParam();
		stringParam.add(new StringParam("http://2"));
		targetDomains.addAnd(stringParam);

		try {
			provider.findPatientsByIdentifier(sourceId, targetDomains);
		} catch (InvalidRequestException e) {
			assertEquals(ChPatientResourceProvider.SOURCE_IDENTIFIER_PATIENT_IDENTIFIER_NOT_FOUND, e.getMessage());
		}

	}

	@Test
	@Description("Test on ihe_pix operation, when the patient system identifiers is null")
	@Severity(SeverityLevel.CRITICAL)
	@Story("ihe_pix operation")
	void testSourceIdentifierSystemNull() {
		TokenParam sourceId = new TokenParam();
		sourceId.setValue("69420");
		StringAndListParam targetDomains = new StringAndListParam();
		StringOrListParam stringParam = new StringOrListParam();
		stringParam.add(new StringParam("http://2"));
		targetDomains.addAnd(stringParam);

		try {
			provider.findPatientsByIdentifier(sourceId, targetDomains);
		} catch (InvalidRequestException e) {
			assertEquals(ChPatientResourceProvider.SOURCE_IDENTIFIER_PATIENT_IDENTIFIER_NOT_FOUND, e.getMessage());
		}

	}

	@Test
	@Description("Test on ihe_pix operation, when the patient source identifiers is null")
	@Severity(SeverityLevel.CRITICAL)
	@Story("ihe_pix operation")
	void testSourceIdentifierValueNull() {
		TokenParam sourceId = new TokenParam();
		sourceId.setSystem("http://1");
		StringAndListParam targetDomains = new StringAndListParam();
		StringOrListParam stringParam = new StringOrListParam();
		stringParam.add(new StringParam("http://2"));
		targetDomains.addAnd(stringParam);

		try {
			provider.findPatientsByIdentifier(sourceId, targetDomains);
		} catch (InvalidRequestException e) {
			assertEquals(ChPatientResourceProvider.SOURCE_IDENTIFIER_ASSIGNING_AUTHORITY_NOT_FOUND, e.getMessage());
		}

	}

	@Test
	@Description("Test on ihe_pix operation, when the patient system identifiers is blank")
	@Severity(SeverityLevel.CRITICAL)
	@Story("ihe_pix operation")
	void testSourceIdentifierSystemBlank() {
		TokenParam sourceId = new TokenParam("", "69420");
		StringAndListParam targetDomains = new StringAndListParam();
		StringOrListParam stringParam = new StringOrListParam();
		stringParam.add(new StringParam("http://2"));
		targetDomains.addAnd(stringParam);

		try {
			provider.findPatientsByIdentifier(sourceId, targetDomains);
		} catch (InvalidRequestException e) {
			assertEquals(ChPatientResourceProvider.SOURCE_IDENTIFIER_ASSIGNING_AUTHORITY_NOT_FOUND, e.getMessage());
		}

	}

	@Test
	@Description("Test on ihe_pix operation, when the patient system identifier hasn't been recognized")
	@Severity(SeverityLevel.CRITICAL)
	@Story("ihe_pix operation")
	void testMalformedSystemIdentifier() {
		TokenParam sourceId = new TokenParam("1:2:3", "69420");
		StringAndListParam targetDomains = new StringAndListParam();
		StringOrListParam stringParam = new StringOrListParam();
		stringParam.add(new StringParam("http://2"));
		targetDomains.addAnd(stringParam);

		try {
			provider.findPatientsByIdentifier(sourceId, targetDomains);
		} catch (ResourceNotFoundException e) {
			assertEquals(ChPatientResourceProvider.SOURCE_IDENTIFIER_PATIENT_IDENTIFIER_NOT_FOUND, e.getMessage());
		}

	}

	@Test
	@Description("Test on ihe_pix operation, when target identifiers is blank")
	@Severity(SeverityLevel.CRITICAL)
	@Story("ihe_pix operation")
	void testTargetIdentifierBlank() {
		TokenParam sourceId = new TokenParam("http::/1", "69420");
		StringAndListParam targetDomains = new StringAndListParam();
		StringOrListParam stringParam = new StringOrListParam();
		stringParam.add(new StringParam(""));
		targetDomains.addAnd(stringParam);

		try {
			provider.findPatientsByIdentifier(sourceId, targetDomains);
		} catch (ForbiddenOperationException e) {
			assertEquals(ChPatientResourceProvider.TARGET_SYSTEM_NOT_FOUND, e.getMessage());
		}
	}

	@Test
	@Description("Test on ihe_pix operation, when no target identifiers is given")
	@Severity(SeverityLevel.CRITICAL)
	@Story("ihe_pix operation")
	void testTargetIdentifierNull() {
		TokenParam sourceId = new TokenParam("http::/1", "69420");
		StringAndListParam targetDomains = new StringAndListParam();
		StringOrListParam stringParam = new StringOrListParam();
		targetDomains.addAnd(stringParam);

		try {
			provider.findPatientsByIdentifier(sourceId, targetDomains);
		} catch (ForbiddenOperationException e) {
			assertEquals(ChPatientResourceProvider.TARGET_SYSTEM_NOT_FOUND, e.getMessage());
		}
	}

	@Test
	@Description("Test on ihe_pix operation, when target identifiers don't have valid cardinality")
	@Severity(SeverityLevel.CRITICAL)
	@Story("ihe_pix operation")
	void testTargetIdentifierMoreThanTwo() {
		TokenParam sourceId = new TokenParam("http::/1", "69420");
		StringAndListParam targetDomains = new StringAndListParam();
		StringOrListParam stringParam = new StringOrListParam();
		stringParam.add(new StringParam("http://2"));
		stringParam.add(new StringParam("http://3"));
		stringParam.add(new StringParam("http://4"));
		targetDomains.addAnd(stringParam);

		try {
			provider.findPatientsByIdentifier(sourceId, targetDomains);
		} catch (ForbiddenOperationException e) {
			assertEquals(ChPatientResourceProvider.TARGET_SYSTEM_NOT_FOUND, e.getMessage());
		}
	}	

	@Test
	@Description("Test on read method, but resource is not found")
	@Severity(SeverityLevel.CRITICAL)
	@Story("ReadMethod")
	void testReadResourceNotFound() {
		IdType id = new IdType();
		id.setValue("1");
		assertThrows(InternalErrorException.class,
				() -> read_provider.read(id));
	}



	@Test
	@Description("Test on read method, but no parameter is given")
	@Severity(SeverityLevel.CRITICAL)
	@Story("ReadMethod")
	void testReadNoIdGiven() {
		IdType id = new IdType();
		id.setValue("");
		try {
			read_provider.read(id);
		} catch (InvalidRequestException e) {
			assertEquals(ChPatientResourceProvider.NO_ID_PROVIDED, e.getMessage());
		} catch (Exception e) {
			fail();
		}
	}

	@Test
	@Description("Test on read method, returns ok")
	@Severity(SeverityLevel.CRITICAL)
	@Story("ReadMethod")
	void testReadFine() {
		IdType id = new IdType();
		id.setValue("3");
		assertEquals("[Arthur]", read_provider.read(id).getName().get(0).getGiven().toString());
	}

	@Test
	@Description("Test on Update method, returns ok")
	@Severity(SeverityLevel.CRITICAL)
	@Story("UpdateMethod")
	void testUpdateFine() {
		IdType id = new IdType();
		Patient iti93Bundle = new Patient();
		iti93Bundle.setId("Plop");
		id.setValue("3");
		Bundle returnedValue = (Bundle) feed_provider.update(id, iti93Bundle).getResource();
		assertEquals("Plop", returnedValue.getEntry().get(0).getResource().getId());
	}

	@Test
	@Description("Test on Update method, returns ok")
	@Severity(SeverityLevel.CRITICAL)
	@Story("UpdateMethod")
	void testUpdateUuidBlank() {
		IdType id = new IdType();
		Patient iti93Bundle = new Patient();
		iti93Bundle.setId("Plop");
		id.setValue("");
		assertThrows(InvalidRequestException.class, () -> feed_provider.update(id, iti93Bundle));
	}

	@Test
	@Description("Test on Update method, returns ok")
	@Severity(SeverityLevel.CRITICAL)
	@Story("UpdateMethod")
	void testUpdatePatientBlank() {
		IdType id = new IdType();
		Patient iti93Bundle = new Patient();
		id.setValue("3");
		assertThrows(InvalidRequestException.class, () -> feed_provider.update(id, iti93Bundle));
	}

	@Test
	@Description("Test on Update method, returns ok")
	@Severity(SeverityLevel.CRITICAL)
	@Story("UpdateMethod")
	void testUpdatePatientFeedThrown() {
		IdType id = new IdType();
		Patient iti93Bundle = new Patient();
		iti93Bundle.setId("Plop");
		id.setValue("42");
		assertThrows(InternalErrorException.class, () -> feed_provider.update(id, iti93Bundle));
	}

	@Test
	@Description("Test on Create method, returns ok")
	@Severity(SeverityLevel.CRITICAL)
	@Story("CreateMethod")
	void testCreateFine() {
		Patient iti93Bundle = new Patient();
		iti93Bundle.setId("PatientPixMFeed");
		Bundle returnedValue = (Bundle) feed_provider.create(iti93Bundle).getResource();
		assertEquals("1", returnedValue.getId());
	}

	@Test
	@Description("Test on Create method, returns ok")
	@Severity(SeverityLevel.CRITICAL)
	@Story("CreateMethod")
	void testCreatePatientBlank() {
		assertThrows(InvalidRequestException.class, () -> feed_provider.create(null));
	}

	@Test
	@Description("Test on Create method, returns nothing")
	@Severity(SeverityLevel.CRITICAL)
	@Story("CreateMethod")
	void testCreatePatientNPE() {
		Patient iti93Bundle = new Patient();
		iti93Bundle.setId("Plop");
		assertThrows(ResourceNotFoundException.class, () -> feed_provider.create(iti93Bundle));
	}

	@Test
	@Description("Test on Create method, returns nothing")
	@Severity(SeverityLevel.CRITICAL)
	@Story("CreateMethod")
	void testCreatePatientFeedThrown() {
		Patient iti93Bundle = new Patient();
		iti93Bundle.setId("throw_me_something");
		assertThrows(InternalErrorException.class, () -> feed_provider.create(iti93Bundle));
	}

	@Test
	@Description("Test on Delete method, returns ok")
	@Severity(SeverityLevel.CRITICAL)
	@Story("DeleteMethod")
	void testDeleteFine() {
		IdType theId = new IdType();
		theId.setValue("PatientPixMFeed");
		Bundle returnedValue = (Bundle) feed_provider.delete(theId).getResource();
		assertEquals("PatientPixMFeed", returnedValue.getId());
	}

	@Test
	@Description("Test on Delete method, returns ok")
	@Severity(SeverityLevel.CRITICAL)
	@Story("DeleteMethod")
	void testDeleteUuidEmpty() {
		IdType theId = new IdType();
		theId.setValue("");
		assertThrows(InvalidRequestException.class, () -> feed_provider.delete(theId));
	}


	@Test
	@Description("Test on Delete method, returns ok")
	@Severity(SeverityLevel.CRITICAL)
	@Story("DeleteMethod")
	void testDeletePatientFeedThrown() {
		IdType theId = new IdType();
		theId.setValue("42");
		assertThrows(InternalErrorException.class, () -> feed_provider.delete(theId));
	}

}