package net.ihe.gazelle.business.provider;

import ca.uhn.fhir.rest.param.StringAndListParam;
import ca.uhn.fhir.rest.param.StringOrListParam;
import ca.uhn.fhir.rest.param.StringParam;
import ca.uhn.fhir.rest.param.TokenParam;
import ca.uhn.fhir.rest.server.exceptions.ForbiddenOperationException;
import ca.uhn.fhir.rest.server.exceptions.InternalErrorException;
import ca.uhn.fhir.rest.server.exceptions.InvalidRequestException;
import ca.uhn.fhir.rest.server.exceptions.ResourceNotFoundException;
import io.qameta.allure.Description;
import io.qameta.allure.Severity;
import io.qameta.allure.SeverityLevel;
import io.qameta.allure.Story;
import net.ihe.gazelle.app.patientregistryapi.application.SearchCrossReferenceException;
import net.ihe.gazelle.application.PatientRegistryXRefSearchClient;
import org.hl7.fhir.r4.model.Identifier;
import org.hl7.fhir.r4.model.Parameters;
import org.hl7.fhir.r4.model.Parameters.ParametersParameterComponent;
import org.junit.jupiter.api.BeforeAll;
import org.junit.jupiter.api.Test;

import static org.junit.jupiter.api.Assertions.assertEquals;

@Story("IHE Patient Provider")
class IhePatientResourceProviderTest {
    private static PatientRegistryXRefSearchClient xRefSearchClientMock = new PatientRegistryXRefSearchClientMock();
    private static final IhePatientResourceProvider provider = new IhePatientResourceProvider(xRefSearchClientMock);

    @BeforeAll
    static void initialized() {
        xRefSearchClientMock = new PatientRegistryXRefSearchClientMock();
    }

    @Test
    @Description("Test on ihe_pix operation, when aiming for a patient on 1 target System, with urn:oid: type system identifiers")
    @Severity(SeverityLevel.CRITICAL)
    @Story("ihe_pix operation")
    void testOkPatient() {
        TokenParam sourceId = new TokenParam("urn:oid:1", "69420");
        StringAndListParam targetDomains = new StringAndListParam();
        StringOrListParam stringParam = new StringOrListParam();
        stringParam.add(new StringParam("urn:oid:2"));
        targetDomains.addAnd(stringParam);

        Parameters response = provider.findPatientsByIdentifier(sourceId, targetDomains);
        ParametersParameterComponent singleResponse = response.getParameter().get(0);
        Identifier r4 = (Identifier) singleResponse.getValue();
        assertEquals(PatientRegistryXRefSearchClientMock.URN_OK, r4.getValue());
        assertEquals(PatientRegistryXRefSearchClientMock.URN_OK, r4.getSystem());

    }

    @Test
    @Description("Test on ihe_pix operation, when aiming for a patient on 1 target System, with http system identifiers")
    @Severity(SeverityLevel.CRITICAL)
    @Story("ihe_pix operation")
    void testOkPatientHttp() {
        TokenParam sourceId = new TokenParam("http://1", "69420");
        StringAndListParam targetDomains = new StringAndListParam();
        StringOrListParam stringParam = new StringOrListParam();
        stringParam.add(new StringParam("http://2"));
        targetDomains.addAnd(stringParam);

        Parameters response = provider.findPatientsByIdentifier(sourceId, targetDomains);
        ParametersParameterComponent singleResponse = response.getParameter().get(0);
        Identifier r4 = (Identifier) singleResponse.getValue();
        assertEquals(PatientRegistryXRefSearchClientMock.HTTP_OK, r4.getValue());
        assertEquals(PatientRegistryXRefSearchClientMock.HTTP_OK, r4.getSystem());

    }

    @Test
    @Description("Test on ihe_pix operation, when the whole patient source identifier is null")
    @Severity(SeverityLevel.CRITICAL)
    @Story("ihe_pix operation")
    void testSourceIdentifierNull() {
        TokenParam sourceId = new TokenParam();
        StringAndListParam targetDomains = new StringAndListParam();
        StringOrListParam stringParam = new StringOrListParam();
        stringParam.add(new StringParam("http://2"));
        targetDomains.addAnd(stringParam);

        try {
            provider.findPatientsByIdentifier(sourceId, targetDomains);
        } catch (InvalidRequestException e) {
            assertEquals(IhePatientResourceProvider.SOURCE_IDENTIFIER_PATIENT_IDENTIFIER_NOT_FOUND, e.getMessage());
        }

    }

    @Test
    @Description("Test on ihe_pix operation, when the patient system identifiers is null")
    @Severity(SeverityLevel.CRITICAL)
    @Story("ihe_pix operation")
    void testSourceIdentifierSystemNull() {
        TokenParam sourceId = new TokenParam();
        sourceId.setValue("69420");
        StringAndListParam targetDomains = new StringAndListParam();
        StringOrListParam stringParam = new StringOrListParam();
        stringParam.add(new StringParam("http://2"));
        targetDomains.addAnd(stringParam);

        try {
            provider.findPatientsByIdentifier(sourceId, targetDomains);
        } catch (InvalidRequestException e) {
            assertEquals(IhePatientResourceProvider.SOURCE_IDENTIFIER_PATIENT_IDENTIFIER_NOT_FOUND, e.getMessage());
        }

    }

    @Test
    @Description("Test on ihe_pix operation, when the patient source identifiers is null")
    @Severity(SeverityLevel.CRITICAL)
    @Story("ihe_pix operation")
    void testSourceIdentifierValueNull() {
        TokenParam sourceId = new TokenParam();
        sourceId.setSystem("http://1");
        StringAndListParam targetDomains = new StringAndListParam();
        StringOrListParam stringParam = new StringOrListParam();
        stringParam.add(new StringParam("http://2"));
        targetDomains.addAnd(stringParam);

        try {
            provider.findPatientsByIdentifier(sourceId, targetDomains);
        } catch (InvalidRequestException e) {
            assertEquals(IhePatientResourceProvider.SOURCE_IDENTIFIER_ASSIGNING_AUTHORITY_NOT_FOUND, e.getMessage());
        }

    }

    @Test
    @Description("Test on ihe_pix operation, when the patient system identifiers is blank")
    @Severity(SeverityLevel.CRITICAL)
    @Story("ihe_pix operation")
    void testSourceIdentifierSystemBlank() {
        TokenParam sourceId = new TokenParam("", "69420");
        StringAndListParam targetDomains = new StringAndListParam();
        StringOrListParam stringParam = new StringOrListParam();
        stringParam.add(new StringParam("http://2"));
        targetDomains.addAnd(stringParam);

        try {
            provider.findPatientsByIdentifier(sourceId, targetDomains);
        } catch (InvalidRequestException e) {
            assertEquals(IhePatientResourceProvider.SOURCE_IDENTIFIER_ASSIGNING_AUTHORITY_NOT_FOUND, e.getMessage());
        }

    }

    @Test
    @Description("Test on ihe_pix operation, when the target system is malformed and not recognized")
    @Severity(SeverityLevel.CRITICAL)
    @Story("ihe_pix operation")
    void testMalformedSystemIdentifier() {
        TokenParam sourceId = new TokenParam("1:2:3", "69420");
        StringAndListParam targetDomains = new StringAndListParam();
        StringOrListParam stringParam = new StringOrListParam();
        stringParam.add(new StringParam("http://2"));
        targetDomains.addAnd(stringParam);

        try {
            provider.findPatientsByIdentifier(sourceId, targetDomains);
        } catch (ResourceNotFoundException e) {
            assertEquals(IhePatientResourceProvider.SOURCE_IDENTIFIER_PATIENT_IDENTIFIER_NOT_FOUND, e.getMessage());
        }

    }

    @Test
    @Description("Test on ihe_pix operation, when the target system is blank")
    @Severity(SeverityLevel.CRITICAL)
    @Story("ihe_pix operation")
    void testTargetIdentifierBlank() {
        TokenParam sourceId = new TokenParam("http::/1", "69420");
        StringAndListParam targetDomains = new StringAndListParam();
        StringOrListParam stringParam = new StringOrListParam();
        stringParam.add(new StringParam(""));
        targetDomains.addAnd(stringParam);

        try {
            provider.findPatientsByIdentifier(sourceId, targetDomains);
        } catch (ForbiddenOperationException e) {
            assertEquals(IhePatientResourceProvider.TARGET_SYSTEM_NOT_FOUND, e.getMessage());
        }

    }

    @Test
    @Description("Test on ihe_pix operation, when the target system is blank")
    @Severity(SeverityLevel.CRITICAL)
    @Story("ihe_pix operation")
    void testSourceIdentifierBlank() {
        TokenParam sourceId = new TokenParam("http::/1", "69420");
        StringAndListParam targetDomains = new StringAndListParam();
        StringOrListParam stringParam = new StringOrListParam();
        stringParam.add(new StringParam(""));
        targetDomains.addAnd(stringParam);

        try {
            provider.findPatientsByIdentifier(null, targetDomains);
        } catch (InvalidRequestException e) {
            assertEquals(IhePatientResourceProvider.SOURCE_IDENTIFIER_PATIENT_IDENTIFIER_NOT_FOUND, e.getMessage());
        }

    }

    @Test
    @Description("Test on ihe_pix operation, when the target system is blank")
    @Severity(SeverityLevel.CRITICAL)
    @Story("ihe_pix operation")
    void testCrossRefException() {
        TokenParam sourceId = new TokenParam("urn:oid:3", "69420");
        StringAndListParam targetDomains = new StringAndListParam();
        StringOrListParam stringParam = new StringOrListParam();
        stringParam.add(new StringParam("urn:oid:3"));
        targetDomains.addAnd(stringParam);

        try {
            provider.findPatientsByIdentifier(sourceId, targetDomains);
        } catch (InternalErrorException e) {
            assertEquals("net.ihe.gazelle.business.provider.PatientResourceProviderException: One of the target domain does not exist", e.getMessage());
        }

    }

}